//  This file is part of the Mussa source distribution.
//  http://mussa.caltech.edu/
//  Contact author: Tristan  De Buysscher, tristan@caltech.edu

// This program and all associated source code files are Copyright (C) 2005
// the California Institute of Technology, Pasadena, CA, 91125 USA.  It is
// under the GNU Public License; please see the included LICENSE.txt
// file for more information, or contact Tristan directly.


//                        ----------------------------------------
//                         ---------- mussa_nway.cc  -----------
//                        ----------------------------------------

#include <boost/filesystem/fstream.hpp>
namespace fs = boost::filesystem;

#include "alg/mussa_callback.hpp"
#include "alg/nway_paths.hpp"
#include "alg/conserved_path.hpp"
#include "mussa_exceptions.hpp"

#include <iostream>
#include <stdexcept>

using namespace std;

NwayPaths::NwayPaths()
{
}

NwayPaths::NwayPaths(const NwayPaths &o)
  : pathz(o.pathz),
    refined_pathz(o.refined_pathz),
    threshold(o.threshold),
    win_size(o.win_size),
    soft_thres(o.soft_thres),
    ent_thres(o.ent_thres),
    c_sequences(o.c_sequences)
{
}

void NwayPaths::clear()
{
  c_sequences.clear();
  pathz.clear();
  refined_pathz.clear();
}

void
NwayPaths::setup(int w, int t)
{
  threshold = t;
  soft_thres = threshold;
  win_size = w;
  clear();

  //cout << "nway: thres = " << threshold
  //     << ", soft threo = " << soft_thres << endl;
}

void
NwayPaths::set_soft_threshold(int sft_thres)
{
  soft_thres = sft_thres;
}

int NwayPaths::get_soft_threshold() const
{
  return soft_thres;
}

int NwayPaths::get_threshold() const
{
  return threshold;
}

int NwayPaths::get_window() const
{
  return win_size;
}

// dumbly goes thru and combines path windows exactly adjacent (ie + 1 index)
// doesn't deal with interleaved adjacency
void
NwayPaths::simple_refine()
{
  // ext_path remembers the first window set in an extending path
  ConservedPath ext_path, new_path;
  list<ConservedPath>::iterator cur_path, next_path;
  list<ConservedPath>::iterator pathz_i;
  int win_ext_len = 0;
  bool extending, end = false;

  refined_pathz.clear();

  //cout << "path number is: " << pathz.size() << endl;
  pathz_i = pathz.begin();
  int path_count = 0;

  // only try to extend when pathz isn't empty.
  if (pathz_i != pathz.end())
  {
    ext_path = *pathz_i;
    ++path_count;

    while(pathz_i != pathz.end())
    {
      // keep track of current path and advance to next path
      cur_path = pathz_i;
      ++pathz_i;
      ++path_count;

      if (pathz_i == pathz.end()) {
        end = true;
        extending = false;
      } else {
        next_path = pathz_i;
        // if node for each seq is equal to the next node+1 then for all
        // sequences then we are extending
        extending = cur_path->nextTo(*next_path);
      }

      if (extending)
      {
        win_ext_len++;
      }
      else
      {
        // add the extend window length as first element and add as refined
        // now that we have the path to extend save it
        new_path = ext_path;
        new_path.extend(win_ext_len);
        refined_pathz.push_back(new_path);
        if (not end) {
          // reset stuff
          win_ext_len = 0;
          ext_path = *next_path;
        }
      }
      if ((path_count % 100) == 0)
        emit progress("refine", path_count-1, pathz.size());
    }
  }
  // this mysterious call tells the dialog box that we're actually done
  emit progress("refine", pathz.size(), pathz.size());
  //cout << "r_path number is: " << refined_pathz.size() << endl;
}

void
NwayPaths::add_path(int threshold, vector<int>& loaded_path)
{
  pathz.push_back(ConservedPath(threshold, 0.0, loaded_path));
}

void
NwayPaths::add_path(ConservedPath loaded_path)
{
  pathz.push_back(loaded_path);
}


void
NwayPaths::save(fs::path save_file_path)
{
  fs::fstream save_file;
  list<ConservedPath >::iterator path_i, paths_end;

  save_file.open(save_file_path, ios::out);

  save_file << "<Mussa type=flp seq_count=" << sequence_count();
  save_file << " win=" << win_size;
  // add a function para new_thres defaults to -1 to later deal with
  // reanalysis with higher thres - if statement whether to record base thres
  // or new thres (ie if -1, then base)
  save_file << " thres=" << threshold << " soft_thres=" << soft_thres << " >\n";

  path_i = refined_pathz.begin();
  paths_end = refined_pathz.end();
  //path_i = pathz.begin();
  //paths_end = pathz.end();
  while (path_i != paths_end)
  {
    ConservedPath& a_path = *path_i;
    //cout << a_path.size() << endl;
    //first entry is the window length of the windows in the path
    save_file << a_path.window_size << ":";
    for(size_type i = 0; i != sequence_count(); ++i)
    {
      save_file << a_path[i];
      if (i != sequence_count())
        save_file << ",";
    }
    save_file << endl;
    ++path_i;
  }

  save_file << "</Mussa>\n";
  save_file.close();
}


NwayPaths::size_type NwayPaths::sequence_count() const
{
  if (refined_pathz.begin() == refined_pathz.end() )
    return 0;
  else
    return refined_pathz.begin()->size();
}

NwayPaths::size_type NwayPaths::size() const
{
  return pathz.size();
}  

void
NwayPaths::load(fs::path load_file_path)
{
  fs::fstream load_file;
  string file_data_line, header_data, data, path_node, path_width;
  int space_split_i, equal_split_i, comma_split_i, colon_split_i;
  vector<int> loaded_path;

  load_file.open(load_file_path, ios::in);

  if (!load_file)
  {
    throw mussa_load_error("Sequence File: " + load_file_path.string() + " not found");
  }
  else
  {
    // get header data
    // grab mussa tag - discard for now...maybe check in future...
    getline(load_file,file_data_line);
    space_split_i = file_data_line.find(" ");
    file_data_line = file_data_line.substr(space_split_i+1);
    // grab type tag - need in future to distinguish between flp and vlp paths
    space_split_i = file_data_line.find(" ");
    file_data_line = file_data_line.substr(space_split_i+1);
    // get species/seq number
    space_split_i = file_data_line.find(" ");
    header_data = file_data_line.substr(0,space_split_i); 
    equal_split_i = header_data.find("=");
    data = file_data_line.substr(equal_split_i+1); 
    unsigned int species_num = atoi (data.c_str());
    file_data_line = file_data_line.substr(space_split_i+1);
    // get window size
    space_split_i = file_data_line.find(" ");
    header_data = file_data_line.substr(0,space_split_i); 
    equal_split_i = header_data.find("=");
    data = file_data_line.substr(equal_split_i+1); 
    win_size = atoi (data.c_str());
    file_data_line = file_data_line.substr(space_split_i+1);
    // get threshold size
    space_split_i = file_data_line.find(" ");
    header_data = file_data_line.substr(0,space_split_i); 
    equal_split_i = header_data.find("=");
    data = file_data_line.substr(equal_split_i+1); 
    threshold = atoi (data.c_str());
    file_data_line = file_data_line.substr(space_split_i+1);
    // get cur_threshold
    //std::cout << "file_data_line: " << file_data_line << "\n";
    //std::cout << "find(\">\"): " << file_data_line.find(">") << "\n";
    if (file_data_line.find(">") != 0)
    {
      space_split_i = file_data_line.find(" ");
      header_data = file_data_line.substr(0,space_split_i);
      equal_split_i =  header_data.find("=");
      data = file_data_line.substr(equal_split_i+1);
      soft_thres = atoi (data.c_str());
      file_data_line = file_data_line.substr(space_split_i+1);
    }
    else
    {
      soft_thres = threshold;
    }
    //std::cout << "nway_soft_thres: " << soft_thres << "\n";
    //cout << "seq_num=" << species_num << " win=" << win_size;
    //cout << " thres=" << threshold << endl;
    
    // clear out the current data
    refined_pathz.clear();
    
    int temp;
    
    getline(load_file,file_data_line);
    while ( (!load_file.eof()) && (file_data_line != "</Mussa>") )
    {
      if (file_data_line != "")
      {
        loaded_path.clear();
        colon_split_i = file_data_line.find(":");
        // whats our window size?
        path_width = file_data_line.substr(0,colon_split_i); 
        file_data_line = file_data_line.substr(colon_split_i+1);
        for(size_type i = 0; i < species_num; i++)
        {
          comma_split_i = file_data_line.find(",");
          path_node = file_data_line.substr(0, comma_split_i); 
          temp = atoi (path_node.c_str());
          loaded_path.push_back(temp);
          file_data_line = file_data_line.substr(comma_split_i+1);
        }
        assert (loaded_path.size() == species_num );
        refined_pathz.push_back(ConservedPath(atoi(path_width.c_str()), 
                                                   threshold, 
                                                   loaded_path));
      }
      getline(load_file,file_data_line);
    }
    load_file.close();
  }
}


void
NwayPaths::path_search(vector<vector<FLPs> > all_comparisons, ConservedPath path, size_type depth)
{
  list<int> new_nodes, trans_check_nodes;
  list<int>::iterator new_nodes_i, new_nodes_end;
  bool trans_check_good;

  new_nodes = all_comparisons[depth - 1][depth].match_locations(path[depth-1]);
  new_nodes_i = new_nodes.begin();
  new_nodes_end = new_nodes.end();
  while(new_nodes_i != new_nodes_end)
  {
    //cout << "    * species " << depth << " node: " << *new_nodes_i << endl;
    // check transitivity with previous nodes in path
    trans_check_good = true;
    for(size_type i = 0; i < depth - 1; i++)
    {
      trans_check_nodes = all_comparisons[i][depth].match_locations(path[i]);
      if ( (trans_check_nodes.end() == find(trans_check_nodes.begin(),
                                            trans_check_nodes.end(),
                                            *new_nodes_i) ) &&
           (trans_check_nodes.end() == find(trans_check_nodes.begin(),
                                           trans_check_nodes.end(),
                                           *new_nodes_i * -1) ) )
        trans_check_good = false;
    }

    if (trans_check_good)
    {
      // this makes sure path nodes are recorded with RC status relative to
      // the base species
      if ( path[depth-1] >= 0)
        path.push_back(*new_nodes_i);
      else
        path.push_back(*new_nodes_i * -1);

      if (depth < all_comparisons.size() - 1)
        path_search(all_comparisons, path, depth + 1);
      else
        pathz.push_back(path);
      path.pop_back();
    } 
    ++new_nodes_i;
  }
}
/* use this if I ever get the friggin seqcomp match lists to sort...
      if (binary_search(trans_check_nodes.begin(), trans_check_nodes.end(), 
                        *new_nodes_i))
*/

void
NwayPaths::find_paths_r(vector<vector<FLPs> > all_comparisons)
{
  ConservedPath path;
  int win_i, window_num;
  list<int> new_nodes;
  list<int>::iterator new_nodes_i, new_nodes_end;

  pathz.clear();
  window_num = all_comparisons[0][1].size();
  // loop thru all windows in first species
  for (win_i = 0; win_i < window_num; win_i++)
  {
    path.clear();
    path.push_back(win_i);
    new_nodes = all_comparisons[0][1].match_locations(path[0]);
    new_nodes_i = new_nodes.begin();
    new_nodes_end = new_nodes.end();
    //if (new_nodes_i != new_nodes_end)
    //cout << "* species 0 node: " << win_i << endl;
    path.push_back(0);
    while(new_nodes_i != new_nodes_end)
    {
      //cout << "  * species 1 node: " << *new_nodes_i << endl;
      path[1] = *new_nodes_i;
      path_search(all_comparisons, path, 2);
      ++new_nodes_i;
    }
  }
}


void
NwayPaths::save_old(fs::path save_file_path)
{
  fs::fstream save_file;
  list<ConservedPath >::iterator path_i, paths_end;
  size_type i;

  save_file.open(save_file_path, ios::app);

  path_i = pathz.begin();
  paths_end = pathz.end();
  while(path_i != paths_end)
  {
    ConservedPath& a_path = *path_i;
    //cout << a_path.size() << endl;
    for(i = 0; i < path_i->size(); ++i)
      save_file << i << "," << a_path[i] << " ";
    save_file << endl;
    ++path_i;
  }
  save_file.close();
}


/*
void
NwayPaths::find_paths(vector<vector<FLPs> > all_comparisons)
{
  int win_i, sp_i;
  vector<list<list<int> > > path_src_tree;
  list<int> new_nodes;
  list<int>::iterator node_i, node_end;
  <list<list<int> > >::iterator branch_i, branch_end;  

  pathz.clear();
  path_src_tree.reserve(all_comparisons.size()- 1);


  // loop thru all windows in first species
  for (win_i = 0; win_i < window_num; win_i++)
  {
    // clear the path search tree
    for(i = 0; i < all_comparisons.size(); i++)
      path_src_tree[i].clear();

    // top level kept empty even tho implicity has one entry of the first
    // species at this window - why bother, adds a little speed

    // get connection list for first species, creating a list of nodes
    // of second species connected to the first species at this window
    new_nodes = all_comparisons[0][1];
    path_src_tree[1].push_back(new_nodes);

    // loop thru rest of species for this window to see if any paths of matches
    // go across all species
    // if path search tree becomes empty, break out of loop, no reason to search further
    sp_i = 1;
    while ((sp_i < all_comparisons.size()) && (path tree not empty))
    {
      branch_i = path_src_tree[1].begin();
      branch_end = path_src_tree[1].end();
      while (branch_i != branch_end)
      {
        node_i = branch_i->begin();
        node_end = branch_i->end();
      }


      // loop over all current nodes
         // get connection list for each node
         // loop over each previous node in list
            // get those nodes connection list
            // intersect previous node connections with current

      ++sp_i;
    }

    // insert any of the paths found into the master list of paths

    // add no paths if tmp_pathz is empty...
  }
}

void NwayPaths::refine()
{
}
*/


void NwayPaths::print(list<vector<int> >& dump_path)
{
  list<vector<int> >::iterator pathz_i;
  vector<int>::iterator path_i;

  cout << "printing list of lists\n";
  for (pathz_i = dump_path.begin(); pathz_i != dump_path.end(); ++pathz_i)
  {
    for (path_i = pathz_i->begin(); path_i != pathz_i->end(); ++path_i)
      cout << *path_i << " ";
    cout << endl;
  }
}
