#include <list>
#include <vector>
#include <sstream>

#include <QStatusBar>
#include <QString>
#include <QMenuBar>

#include "qui/MussaAlignedWindow.hpp"
#include "alg/sequence.hpp"

#include <iostream>
using namespace std;

MussaAlignedWindow::MussaAlignedWindow(MussaRef m,
                                       boost::shared_ptr<QDir> default_dir_,
                                       const set<int>& sel_paths,
                                       SubanalysisWindowRef window, 
                                       QWidget *parent)
  : QMainWindow(parent),
    analysis(m),
    default_dir(default_dir_),
    subanalysis_window(window),
    browser(default_dir),
    pick_align_menu(tr("Choose Alignment")),
    view_align_menu(tr("View Alignment")),
    threshold_widget(0),
    zoom(0),
    alignTB(0)
{
  setupActions();
  connect(&browser, SIGNAL(basepairsCopied(size_t)), 
          this, SLOT(showBasePairsCopied(size_t)));
  browser.setSequences(analysis->sequences(), analysis->colorMapper());
  setSelectedPaths(analysis, sel_paths);
  setAlignment(0);
  double zoom_level = browser.zoomToSequence();

  zoom = new ZoomWidget();
  connect(zoom, SIGNAL(valueChanged(double)), 
          &browser, SLOT(setZoom(double)));
  zoom->setValue(zoom_level);
  
  // Mouse scroll wheel zooming!
  connect(&browser, SIGNAL(mouseWheelZoom(double)),
		  zoom, SLOT(setValue(double)));
  
  computeMatchLines();
  setupMenus();
  setupAlignmentMenus();
  
  setCentralWidget(&browser);

  // Add a threhold widget set to our current threshold and make it readonly 
  threshold_widget = new ThresholdWidget;
  threshold_widget->setRange(analysis->get_threshold(),analysis->get_window());
  threshold_widget->setBasepairThreshold(analysis->get_soft_threshold());
  threshold_widget->setReadOnly(true);
  
  alignTB = new QToolBar();
  alignTB->addWidget(zoom);
  alignTB->addWidget(threshold_widget);
  addToolBar(alignTB);
  
  ostringstream message;
  message << "Selected " << selected_paths.size() << " paths";
  statusBar()->showMessage(message.str().c_str(), 5000);
  browser.updatePosition();
  
  updateTitle();
}

void MussaAlignedWindow::setupActions()
{
  // more cut-n-paste from MussaWindow
  createSubAnalysisAction = new QAction(tr("Add to Subanalysis"), this);
  connect(createSubAnalysisAction, SIGNAL(triggered()), 
          this, SLOT(createSubAnalysis()));
            
  //Save pixel map action
  saveBrowserPixmapAction = new QAction(tr("Save to image..."), this);
  connect(saveBrowserPixmapAction, (SIGNAL(triggered())),
	  &browser, SLOT(promptSaveBrowserPixmap()));
  saveBrowserPixmapAction->setIcon(QIcon(":/icons/image2.png"));
}

void MussaAlignedWindow::setupMenus()
{
  QMenu *newMenu = menuBar()->addMenu(tr("&File"));
  newMenu->addAction(saveBrowserPixmapAction);

  newMenu = menuBar()->addMenu(tr("&Edit"));
  newMenu->addAction(browser.getCopySelectedSequenceAsStringAction());
  newMenu->addAction(browser.getCopySelectedSequenceAsFastaAction());
  newMenu->addAction(createSubAnalysisAction);
  
  // add some extra features to the context menu
  QMenu *popupMenu = browser.getPopupMenu();
  if (popupMenu) {
    popupMenu->addAction(createSubAnalysisAction);
  }  
}

void MussaAlignedWindow::setupAlignmentMenus()
{
  pick_align_menu.clear();
  view_align_menu.clear();
  pick_actions.clear();
  view_actions.clear();

  for(vector<ConservedPath >::iterator pathz_i=selected_paths.begin(); 
      pathz_i != selected_paths.end(); 
      ++pathz_i)
  {
    ConservedPath::path_type normalized_path = pathz_i->normalizedIndexes();
    ostringstream menu_text;
    menu_text << pathz_i->window_size << ":";
    ConservedPath::iterator element_i = normalized_path.begin();
    menu_text << *element_i++;
    for (;
         element_i != normalized_path.end();
         ++element_i)
    {
      menu_text << ", ";
      menu_text << *element_i;
    }
    int index = pathz_i - selected_paths.begin();
    IntAction *pick = new IntAction(QString(menu_text.str().c_str()), index, this);
    connect(pick, SIGNAL(triggered(int)), this, SLOT(setAlignment(int)));
    pick_actions.push_back(pick);
    pick_align_menu.addAction(pick);
    IntAction *view = new IntAction(QString(menu_text.str().c_str()), index, this);
    connect(view, SIGNAL(triggered(int)), this, SLOT(toggleViewAlignment(int)));
    view->setCheckable(true);
    view->setChecked(true);
    view_actions.push_back(view);
    view_align_menu.addAction(view);
  }

  menuBar()->addMenu(&pick_align_menu);
  menuBar()->addMenu(&view_align_menu);
}


void MussaAlignedWindow::setSelectedPaths(MussaRef m, const set<int>& sel_paths)
{
  // sets are sorted
  set<int>::iterator sel_i = sel_paths.begin();
  list<ConservedPath>::const_iterator path_i = m->paths().refined_pathz.begin();
  list<ConservedPath>::const_iterator path_end = m->paths().refined_pathz.end();
  size_t path_size = m->paths().refined_pathz.size();
  size_t pathid=0;

  selected_paths.reserve(sel_paths.size());
  view_paths.reserve(sel_paths.size());
  while (pathid != path_size and sel_i != sel_paths.end())
  {
    assert (*sel_i >= 0);
    size_t sel_pathid = (size_t)(*sel_i);
    if (pathid == sel_pathid) {
      selected_paths.push_back(*path_i);
      view_paths.push_back(true);
      ++pathid;
      ++path_i;
      ++sel_i;
    } else if (pathid < sel_pathid) {
      ++pathid;
      ++path_i;
    } else if (pathid > sel_pathid) {
      ++sel_i;
    }
  }
}

// FIXME: this is a cut-n-paste from MussaWindow, perhaps they should be refactored to
// some shared place
void MussaAlignedWindow::createSubAnalysis()
{
  list<SequenceLocation> result;
  SequenceLocationModel& model = subanalysis_window->getModel();
  browser.copySelectedTracksAsSeqLocation(result);
  for(list<SequenceLocation>::iterator result_itor = result.begin();
      result_itor != result.end();
      ++result_itor)
  {
    model.push_back(*result_itor);
  }

  if (not subanalysis_window->isVisible()) {
    subanalysis_window->show();
  }
}

void MussaAlignedWindow::setAlignment(int alignment_index)
{
  if (selected_paths.size() > 0) {
    browser.centerOnPath(selected_paths[alignment_index].normalizedIndexes());
  }
}

void MussaAlignedWindow::showBasePairsCopied(size_t bp_copied)
{
  QString msg("Copied ");
  QString num;
  num.setNum(bp_copied);
  msg += num + " base pairs";
  statusBar()->showMessage(msg, 5000);
}

void MussaAlignedWindow::toggleViewAlignment(int alignment_index)
{
  view_paths[alignment_index]= not view_paths[alignment_index]; 
  // perhaps it'd be better if we could erase specific sets
  // of matches instead of erasing them all and recomputing them all
  // but this is easier
  computeMatchLines();
}

void MussaAlignedWindow::update()
{
  browser.update();
}

void MussaAlignedWindow::updateTitle()
{
  std::string title("Sequence View: ");
  if (analysis) {
    title += analysis->get_title();
  }
  setWindowTitle(title.c_str());
}

void MussaAlignedWindow::computeMatchLines()
{
  browser.clear_links();
  
  // filter out conserved paths
  list<ConservedPath> filtered_paths;
  vector<ConservedPath>::iterator path_i = selected_paths.begin();
  list<ConservedPath::path_type> result;
  list<vector<bool> > reversed;

  for(vector<ConservedPath>::size_type count = 0; 
      count != selected_paths.size();
      ++count, ++path_i)
  {
    if (view_paths[count]) 
      filtered_paths.push_back(*path_i);
  }
  analysis->createLocalAlignment(filtered_paths.begin(), 
                                 filtered_paths.end(),
                                 result, 
                                 reversed);

  list<ConservedPath::path_type>::const_iterator result_i = result.begin();
  list<vector<bool> >::const_iterator reversed_i = reversed.begin();
  for(int i = 0; i != result.size(); ++i, ++result_i, ++reversed_i)
  {
    // make 1 base long links
    browser.link(*result_i, *reversed_i, 1);
  }
}
